#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Library and command line interface for autots
#
# Copyright (C) 2008-2009 Maximilian Högner
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import pickle
import os, subprocess
import copy

root = os.path.dirname(__file__)

# Localisation
try:
	import gettext
	t = gettext.translation('autots', os.path.join(root,'locale'))
	_ = t.gettext
except Exception, e:
	print str(e)
	_ = lambda x: x

# Exceptions
class Error(Exception):
	def __init__(self, value):
		self.value = value
	def __str__(self):
		return self.value

data = None
path = os.path.join(root,'config.dat')

# Parsing functions

def parse_ip(str):
	sp = str.split('.')

	if len(sp)!=4:
		raise(Error(_("Invalid length of IP address")))

	try:
		t1 = int(sp[0] or 0)
		t2 = int(sp[1] or 0)
		t3 = int(sp[2] or 0)
		t4 = int(sp[3] or 0)
	except:
		raise(Error(_("Invalid characters in IP address.")))

	if t1<0 or t1>254 or t2<0 or t2>254 or t3<0 or t3>254 or t4<0 or t4>254:
		raise(Error(_("Number of IP address out of range.")))

	return "%d.%d.%d.%d" % (t1,t2,t3,t4)

def parse_mac(str):
	sp = str.split(':')

	if len(sp)!=6:
		raise(Error(_("Invalid length of MAC address")))	

	try:
		t1 = int(sp[0] or '0',16)
		t2 = int(sp[1] or '0',16)
		t3 = int(sp[2] or '0',16)
		t4 = int(sp[3] or '0',16)
		t5 = int(sp[4] or '0',16)
		t6 = int(sp[5] or '0',16)
	except:
		raise(Error(_("Invalid characters in MAC address.")))

	if t1<0 or t1>255 or t2<0 or t2>255 or t3<0 or t3>255 or t4<0 or t4>255 or t5<0 or t5>255 or t6<0 or t6>255:
		raise(Error(_("Number of MAC address out of range.")))

	return "%02x:%02x:%02x:%02x:%02x:%02x" % (t1,t2,t3,t4,t5,t6)

def parse_hostname(name):
	if not name or not name.strip():
		raise Error, _("Invalid hostname")
	
	return name.lower().strip()

def parse_roomname(name):
	if not name or not name.strip():
		raise Error, _("Invalid roomname")
	
	return name.strip()


# I/O

def purge():
	globals()['data'] = {
		'rooms':{},
		'config':{
			'hosts':'/etc/hosts',
			'dhcpd.conf':'/etc/ltsp/dhcpd.conf',
			'lts.conf':'/var/lib/tftpboot/ltsp/i386/lts.conf',
			'dhcpd_init_script':'/etc/init.d/dhcp3-server',
			
			# server.py options
			'guest_prefix':'autots-guest-',
			'guest_group':'autots-guest',
			'guest_groups':'fuse',
			'guest_homes':'/opt/autots/guests',
			'guest_model_home':'/opt/autots/model_home',
			'server_interface':'172.16.50.1',
			'server_port':'5000',
			'ltsp4_xsession':'/etc/X11/Xsession'
		}
	}

def load():
	try:
		f = open(globals()['path'], 'r')
	except IOError:
		raise Error(_("Error reading file %s.") % globals()['path'])

	try:
		globals()['data'] = pickle.load(f)
	except:
		raise Error(_("Data file %s invalid.") % globals()['path'])

	f.close()

def save():
	try:
		f = open(path, 'w')
		pickle.dump(globals()['data'], f)
		f.close()
	except:
		raise Error(_("Error saving file %s.") % path)

# config

def update_ltsconf():
	try:
		filename = globals()['data']['config']['lts.conf']
	except:
		raise Error(_("Path for lts.conf not set!"))

	try:
		f = open(filename)
		c = f.read()
		f.close()
	except:
		raise Error(_("Error reading file %s.") % filename)

	try:
		before = c.split("#autots:begin#")[0]
		after = c.split("#autots:end#")[1]
	except:
		before = c
		after = "\n"

	between = ''

	for (room, computer) in computers():
		if len(globals()['data']['rooms'][room]['computers'][computer]['lts.conf'])>0:
			between += "["+computer+"]\n"

			for opt in globals()['data']['rooms'][room]['computers'][computer]['lts.conf']:
				between += "\t"+opt+'\t= "'+globals()['data']['rooms'][room]['computers'][computer]['lts.conf'][opt]+"\"\n"

	try:
		f = open(filename,'w')
		f.writelines(before)
		f.writelines("#autots:begin#\n")
		f.writelines(between+'\n')
		f.writelines("#autots:end#")
		f.writelines(after)
		f.close()
	except IOError:
		raise Error(_("Error saving to file %s.") % filename)

def update_hosts():
	try:
		filename = globals()['data']['config']['hosts']
	except:
		raise Error(_("Path for hosts file not set!"))

	try:
		f = open(filename)
		c = f.read()
		f.close()
	except:
		raise Error(_("Error reading file %s.") % filename)

	try:
		before = c.split("#autots:begin#")[0]
		after = c.split("#autots:end#")[1]
	except:
		before = c
		after = ''

	between = ''

	for (room, computer) in computers():
		between = between+globals()['data']['rooms'][room]['computers'][computer]['IP']+'\t'+computer+'\n'

	try:
		f = open(filename,'w')
		f.writelines(before)
		f.writelines("#autots:begin#\n")
		f.writelines(between+'\n')
		f.writelines("#autots:end#")
		f.writelines(after)
		f.close()
	except IOError:
		raise Error(_("Error saving to file %s.") % filename)

def update_dhcpd():
	try:
		filename = globals()['data']['config']['dhcpd.conf']
	except:
		raise Error(_("Path for dhcpd configuration file not set!"))


	try:
		f = open(filename)
		c = f.read()
		f.close()
	except:
		raise Error(_("Error reading file %s.") % filename)

	try:
		before = c.split("#autots:begin#")[0]
		after = c.split("#autots:end#")[1]
	except:
		raise Error(_("Please write manually the #autots:begin# and #autots:end# entrys to your dhcpd configuration file %s in the right place. Otherwise it can't be updated.") % filename)

	between = ''

	for (room, computer) in computers():
		between = between+'\thost '+computer+'\n\t{\n\t\thardware ethernet '+globals()['data']['rooms'][room]['computers'][computer]['MAC']+';\n\t\tfixed-address '+globals()['data']['rooms'][room]['computers'][computer]['IP']+';\n\t}\n\n'

	try:
		f = open(filename, 'w')
		f.writelines(before)
		f.writelines("#autots:begin#\n")
		f.writelines(between+'\n')
		f.writelines("#autots:end#")
		f.writelines(after)
		f.close()
	except:
		raise Error(_("Error saving to file %s.") % filename)

def dhcpd_restart():
	try:
		command = [ globals()['data']['config']['dhcpd_init_script'], 'restart' ]
	except:
		raise Error(_("Dhcpd restart command not set!"))

	try:
		if not subprocess.call(command)==0:
			raise Exception
	except:
		raise Error(_("Restarting dhcp server by command %s failed.") % command)

# Access functions

def set_config(key,value):
	# prevent user from inventing new config options
	if not key in globals()['data']['config']:
		raise Error(_("Invalid config option \"%s\"") % key)

	globals()['data']['config'][key]=value

def get_config(key=None):
	if key==None:
		r = {}
		for k,v in globals()['data']['config'].iteritems():
			r[k]=v
		return r

	if not key in globals()['data']['config']:
		raise Error(_("Invalid config option \"%s\"") % key)

	return globals()['data']['config'][key]

def rooms():
	return globals()['data']['rooms'].keys()

def room_exists(name):
	room = parse_roomname(name)
	return globals()['data']['rooms'].has_key(room)

def room_create(name):
	room = parse_roomname(name)

	if room_exists(room):
		raise Error(_("Room \"%s\" already exists.") % room)

	globals()['data']['rooms'][room] = {'computers':{}}

def room_remove(name):
	room = parse_roomname(name)

	if not room_exists(room):
		raise Error(_("Room \"%s\" doesn't exist.") % room)

	del globals()['data']['rooms'][room]

def room_rename(old, new):
	old = parse_roomname(old)
	new = parse_roomname(new)

	if not room_exists(old):
		raise Error(_("Room \"%s\" doesn't exist.") % old)

	if old==new:
		return

	if room_exists(new):
		raise Error(_("Room \"%s\" already exists.") % new)

	globals()['data']['rooms'][new] = globals()['data']['rooms'][old]
	room_remove(old)

def room_computers(name):
	room = parse_roomname(name)

	if not room_exists(room):
		raise Error(_("Room \"%s\" doesn't exist.") % room)

	return globals()['data']['rooms'][room]['computers'].keys()

def computers():
	return [(room,computer) for room in globals()['data']['rooms'].keys() for computer in globals()['data']['rooms'][room]['computers'].keys()]

def computer_exists(name):
	computer = parse_hostname(name)

	return computer in [x[1] for x in computers()]

def computer_room(name):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	return filter(lambda x: globals()['data']['rooms'][x]['computers'].has_key(computer), rooms())[0]

def computers_with_property(key, value, but=None):
	return filter(lambda x: x[1]!=but and globals()['data']['rooms'][x[0]]['computers'][x[1]][key]==value, computers())

def computer_create(broom, bcomputer, bip, bmac):
	room = parse_roomname(broom)
	computer = parse_hostname(bcomputer)
	ip = parse_ip(bip)
	mac = parse_mac(bmac)

	if computer_exists(computer):
		raise Error(_("There is already a computer named \"%(computer)s\" in room \"%(room)s\".")
			% {'computer':computer, 'room':computer_room(computer)})

	if not room_exists(room):
		raise Error(_("Room \"%s\" doesn't exist.") % room)


	sip = computers_with_property('IP',ip)
	if sip != []:
		raise Error(_("IP \"%(ip)s\" is occupied by computer \"%(computer)s\" in room \"%(room)s\".")
			% {'ip':ip, 'computer':sip[0][1], 'room':sip[0][0]})

	smac = computers_with_property('MAC',mac)
	if smac != []:
		raise Error(_("MAC \"%(mac)s\" is occupied by computer \"%(computer)s\" in room \"%(room)s\".")
			% {'mac':mac, 'computer':smac[0][1], 'room':smac[0][0]})

	globals()['data']['rooms'][room]['computers'][computer] = {'IP':ip, 'MAC':mac, 'position':(10,10), 'lts.conf':{}}

def computer_get_ip(name):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	return globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['IP']

def computer_set_ip(name, bip):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	ip = parse_ip(bip)
	sip = computers_with_property('IP',ip,name)

	if sip!=[]:
		raise Error(_("IP \"%(ip)s\" is occupied by computer \"%(computer)s\" in room \"%(room)s\".")
			% {'ip':ip, 'computer':sip[0][1], 'room':sip[0][0]})

	globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['IP'] = ip

def computer_get_ltsconf(name, opt=None):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	if not opt:
		return copy.deepcopy(globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['lts.conf'])
	else:
		if not globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['lts.conf'].has_key(opt):
			return ''
		else:
			return globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['lts.conf'][opt]

def computer_set_ltsconf(name, opt, value):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	if value==None:
		if globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['lts.conf'].has_key(opt):
			del globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['lts.conf'][opt]
	else:
		globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['lts.conf'][opt] = value

def computer_set_ltsconf_all(name, conf):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['lts.conf'] = copy.deepcopy(conf)

def computer_get_mac(name):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	return globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['MAC']

def computer_set_mac(name, bmac):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	mac = parse_mac(bmac)
	smac = computers_with_property('MAC',mac,name)

	if smac!=[]:
		raise Error(_("MAC \"%(mac)s\" is occupied by computer \"%(computer)s\" in room \"%(room)s\".")
			% {'mac':mac, 'computer':smac[0][1], 'room':smac[0][0]})

	globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['MAC'] = mac

def computer_get_position(name):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	return globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['position']

def computer_set_position(name, position):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	globals()['data']['rooms'][computer_room(computer)]['computers'][computer]['position'] = position

def computer_remove(name):
	computer = parse_hostname(name)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)

	del globals()['data']['rooms'][computer_room(computer)]['computers'][computer]

def computer_rename(old, new):
	old = parse_hostname(old)
	new = parse_hostname(new)

	if not computer_exists(old):
		raise Error(_("Computer \"%s\" doesn't exist.") % old)

	if old==new:
		return

	room = computer_room(old)

	if computer_exists(new):
		raise Error(_("Computer \"%s\" already exists.") % new)

	globals()['data']['rooms'][room]['computers'][new] = globals()['data']['rooms'][room]['computers'][old]
	del globals()['data']['rooms'][room]['computers'][old]

def computer_move(computer, room):
	computer = parse_hostname(computer)
	room = parse_roomname(room)

	if not computer_exists(computer):
		raise Error(_("Computer \"%s\" doesn't exist.") % computer)
	
	if not room_exists(room):
		raise Error(_("Room \"%s\" doesn't exist.") % room)
	
	oldroom = computer_room(computer)

	if oldroom==room:
		return

	globals()['data']['rooms'][room]['computers'][computer] = globals()['data']['rooms'][oldroom]['computers'][computer]
	del globals()['data']['rooms'][oldroom]['computers'][computer]

# Main

if __name__ == '__main__':
	import sys
	import optparse
	
	optparse._ = _

	# workaround for UnicodeDecodeError
	_backup=_
	def _(text):
		return unicode(_backup(text),"utf-8")

	def ask(question, force):
		if force:
			return True

		print question
		print _("y/N")
		input=sys.stdin.readline()
		input=input.strip()
		if input.lower()==_("y"):
			return True
		else:
			return False

	optParser = optparse.OptionParser(
		usage=_("%prog [-fh] [-parsSuU] [-RCciml] [OPTIONS]"),
		description=_("A tool to manage ltsp clients with static IPs")
	)

	optParser.add_option( "-f", "--force",
	action="store_true", dest="force",
	help=_("don't ask before removing"))

	optParser.add_option( "-s", "--show",
	action="store_true", dest="show",
	help=_("show the data"))
	optParser.add_option( "-p", "--purge",
	action="store_true", dest="purge",
	help=_("purge all data and create a new database"))
	optParser.add_option( "-a", "--add",
	action="store_true", dest="add",
	help=_("add a room or computer"))
	optParser.add_option( "-r", "--remove",
	action="store_true", dest="remove",
	help=_("remove a room or computer"))
	optParser.add_option( "-S", "--set",
	action="store_true", dest="set",
	help=_("configure/rename/move computers and rooms"))
	optParser.add_option( "-U", "--unset",
	action="store_true", dest="unset",
	help=_("unset an option"))
	optParser.add_option( "-u", "--update",
	action="store_true", dest="update",
	help=_("update configuration files"))

	optParser.add_option( "-R", "--room",
	action="store_true", dest="room",
	help=_("apply action on room"))
	optParser.add_option( "-C", "--computer",
	action="store_true", dest="computer",
	help=_("apply action on computer"))

	optParser.add_option( "-c", "--config",
	action="store_true", dest="config",
	help=_("apply action on configuration"))

	optParser.add_option( "-i", "--ip",
	action="store_true", dest="ip",
	help=_("specify a IP address"))
	optParser.add_option( "-m", "--mac",
	action="store_true", dest="mac",
	help=_("specify a MAC address"))

	optParser.add_option( "-l", "--ltsconf",
	action="store_true", dest="ltsconf",
	help=_("specify a lts.conf option"))

	(options, args) = optParser.parse_args()

	if options.purge:
		if ask(_("Do you really want to purge your configuration?"), options.force):
			purge()
		else:
			sys.exit(0)
	else:
		try:
			load()
		except Error, e:
			print str(e)
			print _("Try creating a new configuration by passing \"--purge\" to this program.")
			sys.exit(1)

	if options.purge:
		pass
	elif options.show:
		if options.computer:
			if len(args)<1:
				print _("please specify a computer name")
				sys.exit(1)

			try:
				pcomputer = parse_hostname(args[0])
				ip = computer_get_ip(pcomputer)
				mac = computer_get_mac(pcomputer)
				ltsconf = computer_get_ltsconf(pcomputer)
			except Error,e:
				print str(e)
				sys.exit(1)
			
			print _("name:\t")+pcomputer
			print _("IP:\t")+ip
			print _("MAC:\t")+mac
			print
			print _("lts.conf options:")

			for i in ltsconf:
				print "\t" + i + " = " + ltsconf[i]
		elif options.room:
			if len(args)<1:
				print _("please specify a room name")
				sys.exit(1)

			try:
				room = parse_roomname(args[0])
				lcomputers = room_computers(room)
			except Error,e:
				print str(e)
				sys.exit(1)

			print _("room %(room)s: (%(computers)d total)") % {'room':room, 'computers':len(lcomputers)}

			for computer in lcomputers:
				print "\t%s\t%s\t%s" % (computer, computer_get_ip(computer), computer_get_mac(computer))
		elif options.config:
			if len(args)<1:
				for key,value in get_config().iteritems():
					print key+": "+value
			else:
				try:
					print get_config(args[0])
				except Error,e:
					print str(e)
					sys.exit(1)
		else:
			for room in rooms():
				lcomputers = room_computers(room)

				print _("room %(room)s: (%(computers)d total)") % {'room':room, 'computers':len(lcomputers)}
			
				for computer in lcomputers:
					print "\t%s\t%s\t%s" % (computer, computer_get_ip(computer),
						computer_get_mac(computer))

	elif options.add:
		if options.computer:
			if len(args)<4:
				print _("please specify a room name, computer name, IP adress and MAC address")
				sys.exit(1)

			try:
				computer_create(args[0], args[1], args[2], args[3])
			except Error, e:
				print str(e)
				sys.exit(1)
		elif options.room:
			if len(args)<1:
				print _("please specify a room name")
				sys.exit(1)

			try:
				room_create(args[0])
			except Error, e:
				print str(e)
				sys.exit(1)
		else:
			print _("please add -R or -C")
			sys.exit(1)
	elif options.remove:
		if options.computer:
			if len(args)<1:
				print _("please specify a computer name")
				sys.exit(1)

			try:
				computer = parse_hostname(args[0])
				if ask(_("Do you really want to remove computer %s?") % computer, options.force):
					computer_remove(computer)
			except Error, e:
				print str(e)
				sys.exit(1)
		elif options.room:
			if len(args)<1:
				print _("please specify a room name")
				sys.exit(1)

			try:
				room = parse_roomname(args[0])
				if ask(_("Do you really want to remove room %s?") % room, options.force):
					room_remove(args[0])
			except Error, e:
				print str(e)
				sys.exit(1)
		else:
			print _("please add -R or -C")
			sys.exit(1)
	elif options.unset and options.ltsconf:
		if len(args)<2:
			print _("please specify a computer name and a lts.conf variable to unset")
			sys.exit(1)
		else:
			try:
				computer_set_ltsconf(args[0], args[1], None)
			except Error, e:
				print str(e)
				sys.exit(1)
	elif options.set:
		if options.ltsconf:
			if len(args)<3:
				print _("please specify a computer name, a lts.conf variable and its value")
				sys.exit(1)

			try:
				computer_set_ltsconf(args[0], args[1], args[2])
			except Error, e:
				print str(e)
				sys.exit(1)
		elif options.ip:
			if len(args)<2:
				print _("please specify a computer name and a IP address")
				sys.exit(1)

			try:
				computer_set_ip(args[0], args[1])
			except Error, e:
				print str(e)
				sys.exit(1)
		elif options.mac:
			if len(args)<2:
				print _("please specify a computer name and a MAC address")
				sys.exit(1)

			try:
				computer_set_mac(args[0], args[1])
			except Error, e:
				print str(e)
				sys.exit(1)
		elif options.room and options.computer:
			if len(args)<2:
				print _("please specify computer name and room name")
				sys.exit(1)

			try:
				computer_move(args[0], args[1])
			except Error, e:
				print str(e)
				sys.exit(1)
			
		elif options.room:
			if len(args)<2:
				print _("please specify old and new room name")
				sys.exit(1)

			try:
				room_rename(args[0], args[1])
			except Error, e:
				print str(e)
				sys.exit(1)
		elif options.computer:
			if len(args)<2:
				print _("please specify old and new computer name")
				sys.exit(1)

			try:
				computer_rename(args[0], args[1])
			except Error, e:
				print str(e)
				sys.exit(1)
		elif options.config:
			if len(args)<2:
				print _("please specify config option and value")
			else:
				try:
					set_config(args[0], args[1])
				except Error,e:
					print str(e)
					sys.exit(1)
			

		else:
			print _("add -i or -m to set a computer's IP or MAC address")
			print _("add -R or -C to rename a room or computer")
			print _("add -RC to move a computer to another room")
			print _("add -c to set a configuration option")
			sys.exit(1)
	elif options.update:
		try:
			update_hosts()
			update_dhcpd()
			update_ltsconf()
		except Error, e:
			print str(e)
			sys.exit(1)

		if ask(_("Do you want the dhcp server to be restarted now? (necessary to apply changes)"), options.force):
			try:
				dhcpd_restart()
			except Error,e:
				print str(e)
				sys.exit(1)

	else:
		optParser.print_help()
		sys.exit(0)
	
	try:
		save()
	except Error,e:
		print str(e)
		sys.exit(1)
